(*
   This file is part of The Construct - a graphical user interface
   builder for ML::AbstractUI written in Standard ML.

   Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen)

   E-mail: anoq@HardcoreProcessing.com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing <sales@HardcoreProcessing.com>
   for more information.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(*require "$.collections.simplesets";*)
(*require "$.gtwinnt.dgpcommon";*)
(*require "$.gtwinnt.dgnodes";*)
(*require "$.gtwinnt.dgshorthand";*)
(*require "$.mlabstractuiwinnt.winregistered";*)
(*require "$.mlabstractuiwinnt.lfnmisc";*)
(*require "$.mlabstractuiwinnt.lfngeometry";*)
(*require "$.mlabstractuiwinnt.lfnmiscnotify";*)
(*require "$.mlabstractuiwinnt.lfnsimpletext";*)
(*require "$.basis.__list";*)
(*require "protocolmanip";*)

signature PROTOCOL_ACCESSOR_CONTR =
  sig
    type t = {access : DGNodes.t -> unit,
              clear : unit -> unit,
              getAccessed : unit -> DGNodes.t option,
              checkBinding : DGNodes.t -> (bool * string),
              addBindDesc : DGNodes.t -> string -> unit}
  
    val createController : DGNodes.t ->
                           {protocolAccessor : DGNodes.t} ->
                           t
  end

functor FProtocolAccessorContr() :> PROTOCOL_ACCESSOR_CONTR =
  struct
    open DGShorthand

    type t = {access : DGNodes.t -> unit,
              clear : unit -> unit,
              getAccessed : unit -> DGNodes.t option,
              checkBinding : DGNodes.t -> (bool * string),
              addBindDesc : DGNodes.t -> string -> unit}

    (* FIXME: Put this in a general place... *)
    (* From L.C. Paulson's ML for the Working Programmer p. 110 *)
    fun qsort _ [] = []
      | qsort _ [x] = [x]
      | qsort lessThanEqF (a::bs) =
          let
            fun partition (left, right, []) =
                  (qsort lessThanEqF left) @
                     (a :: qsort lessThanEqF right)
              | partition (left, right, x :: xs) =
                  if lessThanEqF (x, a) then
                    partition (x::left, right, xs)
                  else
                    partition (left, x::right, xs)
          in
            partition ([], [], bs)
          end

    fun createController masterNode {protocolAccessor} =
          let
            val masterCh = masterNode +: "Children"
            val winCh = protocolAccessor +: "Children"
            
            val protocolSlots = ref nil
            val slotHeight = 18

            fun checkBinding node =
                  let
                    (* This line may raise an exception
                       which is caught below *)
                    val bindDescNode = hd (node +# "BindingName")
                              
                    val name = PString.read
                                 (DGPString.p bindDescNode)
                  in
                    (true, name)
                  end
                    handle _ => (false, "")

            fun addValueBinding cnt =
                  let
                    val binder = Reg.createA "ToggleButton"
                    val _ = winCh ~+ binder
                    val _ = LFNGeometry.setPosY binder (slotHeight + cnt * slotHeight)
                    val _ = LFNGeometry.setSizeX binder slotHeight
                    val _ = LFNMisc.createControl binder

                    val bindName = Reg.createA "TextField"
                    val _ = winCh ~+ bindName
                    val _ = LFNGeometry.setPosY bindName (slotHeight + cnt * slotHeight)
                    val _ = LFNGeometry.setPosX bindName slotHeight
                    val _ = LFNGeometry.setSizeX bindName 182
                    val _ = LFNMisc.createControl bindName
                  in
                    {binder = binder,
                     bindName = bindName}
                  end

            fun addEditor cnt =
                  let
                    val nameLabel = Reg.createA "TextLabel"
                    val _ = winCh ~+ nameLabel
                    val _ = LFNGeometry.setPosX nameLabel 200
                    val _ = LFNGeometry.setPosY nameLabel (slotHeight + cnt * slotHeight)
                    val _ = LFNMisc.createControl nameLabel

                    val editField = Reg.createA "TextField"
                    val _ = winCh ~+ editField
                    val _ = LFNGeometry.setPosY editField (slotHeight + cnt * slotHeight)
                    val _ = LFNGeometry.setPosX editField 300
                    val _ = LFNGeometry.setSizeX editField 200
                    val _ = LFNMisc.createControl editField
                  in
                    {nameLabel = nameLabel,
                     editField = editField}
                  end

            fun addProtocolSlot cnt =
                  let
                    val {binder, bindName} = addValueBinding cnt
                    val {nameLabel, editField} = addEditor cnt
                  in
                    {binder = binder, bindName = bindName,
                     nameLabel = nameLabel, editField = editField,
                     changeID = ref NONE,
                     focusChangeID = ref NONE,
                     toggleID = ref NONE,
                     bindNameChangeID = ref NONE}
                  end

            fun setSlotVisibility value
                                  {binder, bindName,
                                   nameLabel, editField,
                                   changeID, focusChangeID,
                                   toggleID, bindNameChangeID} =
                  let
                    val _ = LFNMisc.setVisible binder value
                    val _ = LFNMisc.setVisible bindName value
                    val _ = LFNMisc.setVisible nameLabel value
                    val _ = LFNMisc.setVisible editField value
                  in
                    ()
                  end

            fun strAddBindDesc graphNode name =
                  let
                    val bindDescProt = DGPString.add graphNode
                                         "BindingName"
                                         (name, graphNode)
                  in
                    ()
                  end

            fun addBindDesc graphNode bindName =
                  let
                    val name = PString.read
                                 (DGPString.p
                                    (bindName +: "EditableText"))
                  in
                    strAddBindDesc graphNode name
                  end

            fun removeBindDesc graphNode =
                  let
                    (* This line may raise an exception
                       which is caught below *)
                    val bindDescNode = hd (graphNode +# "BindingName")
                  in
                    DGNodes.disconnect graphNode bindDescNode
                  end
                    handle _ => ()

            fun removeSlotNotifier (slot as {changeID, focusChangeID,
                                             toggleID, bindNameChangeID,
                                             binder,
                                             bindName, nameLabel,
                                             editField}) =
                  (case !changeID of
                     NONE => ()
                   | SOME(theNode, theID) =>
                       LFNMiscNotify.removeEditableTextChangedNotifier theNode theID;
                   changeID := NONE;
                   case !focusChangeID of
                     NONE => ()
                   | SOME(theNode, theID) =>
                       LFNMiscNotify.removeFocusChangedNotifier theNode theID;
                   focusChangeID := NONE;
                   case !toggleID of
                     NONE => ()
                   | SOME(theNode, theID) =>
                       LFNMiscNotify.removeToggleValueChangedNotifier theNode theID;
                   toggleID := NONE;
                   case !bindNameChangeID of
                     NONE => ()
                   | SOME(theNode, theID) =>
                       LFNMiscNotify.removeEditableTextChangedNotifier theNode theID;
                   bindNameChangeID := NONE)

            fun clearSlot slot =
                  (removeSlotNotifier slot;
                   setSlotVisibility false slot)

            fun clearSlots () =
                  List.app clearSlot (!protocolSlots)

            fun setupSlot node (slot as
                                {binder, bindName,
                                 nameLabel, editField,
                                 changeID, focusChangeID,
                                 toggleID, bindNameChangeID}) =
                  let
                    val typeName = DGNodes.typeName node
                    val _ = removeSlotNotifier slot
                    val _ = LFNSimpleText.setText nameLabel typeName

                    fun onBindNameEdit sender value =
                          let
                            (* This line may raise an exception
                               which is caught below *)
                            val bindDescNode = hd (node +# "BindingName")
                          in
                            PString.write (DGPString.p bindDescNode) value
                          end
                            handle _ => ()

                    fun onToggle sender value =
                          case value of
                            true =>
                              (addBindDesc node bindName;
                               PBool.write (DGPBool.p (bindName +: "HasFocus")) true)
                          | false =>
                              (removeBindDesc node)
                  in
                    case ProtocolManip.getProtocol
                           (DGNodes.protocolName node) of
                      NONE =>
                        (LFNSimpleText.setEditableText editField
                           ("N/A: " ^ (DGNodes.protocolName node));
                         changeID := NONE;
                         focusChangeID := NONE;
                         toggleID := NONE;
                         bindNameChangeID := NONE;
                         PBool.write
                           (DGPBool.p (binder +: "ToggleValue")) false;
                         LFNSimpleText.setEditableText  bindName "")
                    | SOME({readAsString, writeAsString, saveSML}) =>
                        let
                          val (hasBinding, theValueOfIt) =
                                 checkBinding node
                          val _ = PBool.write
                                    (DGPBool.p (binder +: "ToggleValue"))
                                    hasBinding
                          val _ = LFNSimpleText.setEditableText
                                    bindName theValueOfIt
                        in
                          (LFNSimpleText.setEditableText editField (readAsString node);
                           changeID := SOME(
                             LFNMiscNotify.addEditableTextChangedNotifier
                               editField
                               (fn _ => fn value => writeAsString (node, value)));
                           focusChangeID := SOME(
                             LFNMiscNotify.addFocusChangedNotifier
                               editField
                               (fn _ => fn value => if value = false then
                                                      LFNSimpleText.setEditableText
                                                        editField (readAsString node)
                                                    else ()));
                           toggleID := SOME(
                             LFNMiscNotify.addToggleValueChangedNotifier
                               binder onToggle);
                           bindNameChangeID := SOME(
                             LFNMiscNotify.addEditableTextChangedNotifier
                               bindName onBindNameEdit))
                        end
                  end

            val lastAccessed = ref NONE

            (* FIXME: Also in PrettyPrinter... *)
            val excludeList = ["Create", "Children",
                               "TCConstructed", "BindingName"]
            
            fun access control =
                  let
                    val _ = lastAccessed := SOME(control)
                    
                    (* Lazy creation of slots as they are needed *)
                    fun setupSlots nil nil _ = nil
                      | setupSlots nil (slot::slots) cnt =
                          (clearSlot slot;
                           slot::setupSlots nil slots (cnt + 1))
                      | setupSlots (node::nodes) nil cnt =
                          let
                            val newSlot = addProtocolSlot cnt
                            val _ = setupSlot node newSlot
                          in
                            newSlot::setupSlots nodes nil (cnt + 1)
                          end
                      | setupSlots (node::nodes) (slot::slots) cnt =
                          let
                            val _ = setSlotVisibility true slot
                            val _ = setupSlot node slot
                          in
                            slot::setupSlots nodes slots (cnt + 1)
                          end

                    val connected = DGNodes.connected control
                    fun filterF node =
                          not (SimpleSets.contains
                                 (DGNodes.typeName node)
                                 excludeList)
                    fun lessThanEqF (node1, node2) =
                          DGNodes.typeName node1 <=
                            DGNodes.typeName node2
                          
                    val connected =
                          qsort lessThanEqF
                            (List.filter filterF connected)
                    
                    val accessNodes = control::connected

                    val _ = protocolSlots := setupSlots
                              accessNodes (!protocolSlots) 0

                    val _ = LFNGeometry.setSizeY
                              protocolAccessor
                              (slotHeight + (length accessNodes) * slotHeight)
                  in
                    () 
                  end

            fun clear () =
                  (lastAccessed := NONE;
                   clearSlots ();
                   LFNGeometry.setSizeY protocolAccessor slotHeight)

            fun getAccessed () =
                  !lastAccessed

            val _ = clear ()
          in
            {access = access,
             clear = clear,
             getAccessed = getAccessed,
             checkBinding = checkBinding,
             addBindDesc = strAddBindDesc}
          end
  end

structure ProtocolAccessorContr = FProtocolAccessorContr()
