(*
   This file is part of The Construct - a graphical user interface
   builder for ML::AbstractUI written in Standard ML.

   Copyright (C) 1999 ANOQ of the Sun (alias Johnny Andersen)

   E-mail: anoq@HardcoreProcessing.com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   If you are interested in a warranty or commercial support for this
   software, contact Hardcore Processing <sales@HardcoreProcessing.com>
   for more information.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(*require "$.collectionswinnt.dictionary";*)
(*require "$.gtwinnt.dgpcommon";*)
(*require "$.basis.__int";*)
(*require "$.gtwinnt.dgnodes";*)
(*require "$.mlabstractui.pnotifiers";*)
(*require "$.mlabstractui.uiprotocols";*)

functor FProtocolManip() =
  struct
    val funcDict = ref (StringDict.empty :
                          int StringDict.t)

    fun addFunction name args =
          (funcDict :=
             StringDict.insert
               (!funcDict) name args)

    fun functionArgs name =
          case StringDict.lookup
                 (!funcDict) name of
            NONE => 0 (* It is not a function... *)
          | SOME(args) => args
  
    val _ = addFunction "Reg.createA" 1
  
    type registeredP =
           {readAsString : DGNodes.t -> string,
            writeAsString : (DGNodes.t * string) -> unit,
            saveSML : ((string -> unit) * (unit -> unit)) ->
                      (DGNodes.t * string * string) -> unit}
  
    val protDict = ref (StringDict.empty :
                          registeredP StringDict.t)

    fun addProtocol {readAsString, writeAsString, typeName, saveSML} =
          (protDict :=
             StringDict.insert
               (!protDict) typeName
               {readAsString = readAsString,
                writeAsString = writeAsString,
                saveSML = saveSML};
           addFunction ("DGP" ^ typeName ^ ".p") 1;
           addFunction ("P" ^ typeName ^ ".write") 2)

    fun getProtocol typeName =
          StringDict.lookup
            (!protDict) typeName

    (* PString protocol *)
    fun readPString node =
          PString.read (DGPString.p node)

    fun savePString (outF, outLF) (node, binding, protBinding) =
          (outF ("val " ^ protBinding);
           outF " = DGPString.p ";
           outF binding; outLF ();
           outF ("val _ = PString.write " ^ protBinding ^ " \"");
           outF (readPString node); outF "\""; outLF ())

    val _ = addProtocol
              {typeName = "String",
               readAsString = readPString,
               writeAsString = fn (node, value) =>
                                 PString.write (DGPString.p node) value,
               saveSML = savePString}

    (* PBool protocol *)
    fun readPBool node =
          case PBool.read (DGPBool.p node) of
            true => "true"
          | false => "false"

    fun writePBool (node, value) =
          case value of
            "true" => PBool.write (DGPBool.p node) true
          | "false" => PBool.write (DGPBool.p node) false
          | _ => ()

    fun savePBool (outF, outLF) (node, binding, protBinding) =
          (outF ("val " ^ protBinding ^ " = DGPBool.p ");
           outF binding; outLF ();
           outF ("val _ = PBool.write " ^ protBinding ^ " ");
           outF (readPBool node); outLF ())

    val _ = addProtocol
              {typeName = "Bool",
               readAsString = readPBool,
               writeAsString = writePBool,
               saveSML = savePBool}


    (* PInt protocol *)
    fun readPInt node =
          Int.toString(PInt.read (DGPInt.p node))

    fun savePInt (outF, outLF) (node, binding, protBinding) =
          (outF ("val " ^ protBinding ^ " = DGPInt.p ");
           outF binding; outLF ();
           outF ("val _ = PInt.write " ^ protBinding ^ " ");
           outF (readPInt node); outLF ())

    val _ = addProtocol
              {typeName = "Int",
               readAsString = readPInt,
               writeAsString = fn (node, value) =>
                                 (case Int.fromString value of
                                    NONE => ()
                                  | SOME(anInt) =>
                                      PInt.write
                                        (DGPInt.p node)
                                        anInt),
               saveSML = savePInt}

    (* PNotify protocol *)
    fun readPNotify node =
          "(only for bindings)"

    fun savePNotify (outF, outLF) (node, binding, protBinding) =
          (outF ("val " ^ protBinding ^ " = DGPNotify.p ");
           outF binding; outLF ())

    val _ = addProtocol
              {typeName = "Notify",
               readAsString = readPNotify,
               writeAsString = fn (node, value) => (),
               saveSML = savePNotify}

    (* PPointerButtonNotify protocol *)
    fun readPPointerButtonNotify node =
          "(only for bindings)"

    fun savePPointerButtonNotify (outF, outLF) (node, binding, protBinding) =
          (outF ("val " ^ protBinding ^ " = DGPPointerButtonNotify.p ");
           outF binding; outLF ())

    val _ = addProtocol
              {typeName = "PointerButtonNotify",
               readAsString = readPPointerButtonNotify,
               writeAsString = fn (node, value) => (),
               saveSML = savePPointerButtonNotify}

    (* PStrings protocol *)
    fun readPStrings node =
          "(only for bindings)"

    fun savePStrings (outF, outLF) (node, binding, protBinding) =
          (outF ("val " ^ protBinding ^ " = DGPStrings.p ");
           outF binding; outLF ())

    val _ = addProtocol
              {typeName = "Strings",
               readAsString = readPStrings,
               writeAsString = fn (node, value) => (),
               saveSML = savePStrings}

    (* PItemSelections protocol *)
    fun readPItemSelections node =
          "(only for bindings)"

    fun savePItemSelections (outF, outLF) (node, binding, protBinding) =
          (outF ("val " ^ protBinding ^ " = DGPItemSelections.p ");
           outF binding; outLF ())

    val _ = addProtocol
              {typeName = "ItemSelections",
               readAsString = readPItemSelections,
               writeAsString = fn (node, value) => (),
               saveSML = savePItemSelections}
  end

structure ProtocolManip = FProtocolManip()
